//========================================================
//HI-MD Renderer Application
//
//Written by Mark Collier.
//http://www.marcnetsystem.co.uk/
//
//This program is freeware and may be distributed in part or 
//in whole as long as a the author is credited clearly some way in 
//both the source code and in some visible way when the program 
//is running (eg:in a help screen or version info screen ).

//This notice must also appear in any source file that contains
//any part of the code in this file. 

//I (the author) take no responsability for the loss, transfer, corruption
//duplication, modification, or addition of legal or illegal data in any way. 
//
//Thanks for reading

//MarC
//=========================================================

#include <windows.h>
#include <stdio.h>
#include <conio.h>
#include <strmif.h>
#include <uuids.h>
#include <vfw.h>
#include <dsound.h>
#include <control.h>
#include <evcode.h> 

//==================Macros================================
#define RELEASE(X) if(X != NULL){X->Release();X=NULL;}
#define WriteLogCR(text) WriteLog(text,TRUE)

//=====================Types================================

typedef struct _WAVSTRUCT
{
	char id[4];
	DWORD dwLen;
	char wave_id[4];
}WAVSTRUCT, *LPWAVSTRUCT;

typedef struct _WAVCHUNK
{
	char chunkid[4];
	DWORD dwChunkDataLen;

}WAVCHUNK, *LPWAVCHUNK;

typedef struct _WAVFORMAT
{
	WORD wFormatTag;
	WORD wChannels;
	DWORD dwSampleRate;
	DWORD dwBufferEst;
	WORD wBlockAlign;
	WORD wBitsPerSample;

}WAVFORMAT, *LPWAVFORMAT;

typedef struct _CUTINFO
{
	LPBYTE lpbtCutFrom;
	DWORD dwCutLen;

}CUTINFO, *LPCUTINFO;

//=================Globals=================================
static HANDLE hLogFile = INVALID_HANDLE_VALUE;
static char infile[MAX_PATH];
static char tempfile[MAX_PATH];
static char destfile[MAX_PATH];
static BOOL bDebug = FALSE;
static BOOL bKeepTmpFile = FALSE;
static DWORD dwBlockSize = 10;
static DWORD dwAllowedDiff = 15;
static BOOL bUseAvging = TRUE;
//====================Logging functions==========================
static void WriteLog(LPSTR text,BOOL bCREnd)
{
	DWORD dwWrit;

	if(hLogFile == INVALID_HANDLE_VALUE)
	{
		printf(text);
		if(bCREnd)
		{
			printf("\n");
		}
		return;
	}

	WriteFile(hLogFile,text,strlen(text),&dwWrit,NULL);
	printf(text);
	if(bCREnd)
	{
		WriteFile(hLogFile,"\r\n",2,&dwWrit,NULL);
		printf("\n");
	}
}

static void WriteLogNum(LPSTR text,DWORD dwValue,BOOL bCREnd)
{
	LPSTR buffer = (LPSTR)LocalAlloc(LPTR,strlen(text) + 500);

	if(buffer != NULL)
	{
		sprintf(buffer,"%s - [0x%x = %u] ",text,dwValue,dwValue);
		WriteLog(buffer,bCREnd);
	}

	LocalFree(buffer);
}

static void WriteLogNumLongLong(LPSTR text,LONGLONG llValue,BOOL bCREnd)
{
	LPSTR buffer = (LPSTR)LocalAlloc(LPTR,strlen(text) + 500);

	if(buffer != NULL)
	{
		sprintf(buffer,"%s - [0x%I64x = %I64u]",text,llValue,llValue);
		WriteLog(buffer,bCREnd);
	}

	LocalFree(buffer);
}

static void WriteLogNumDouble(LPSTR text,double dValue,BOOL bCREnd)
{
	LPSTR buffer = (LPSTR)LocalAlloc(LPTR,strlen(text) + 500);

	if(buffer != NULL)
	{
		sprintf(buffer,"%s - [%g]",text,dValue);
		WriteLog(buffer,bCREnd);
	}

	LocalFree(buffer);
}

static void CreateLogFile(LPSTR name)
{
	char buffer[MAX_PATH];
	
	sprintf(buffer,"%s.log",name);

	hLogFile = CreateFile(buffer,GENERIC_WRITE,FILE_SHARE_READ,NULL,CREATE_ALWAYS,0,NULL);
}
	

///=============================DirectShow utility functions=================================

static int DecodeRet(long lret)
{
	switch(lret)
	{
		case S_OK:
			WriteLogCR("Success !");
			return 0;
			break;
		case REGDB_E_CLASSNOTREG:
			WriteLogCR("Error: Class not registered");
			return -1;
			break;
		case CLASS_E_NOAGGREGATION :
			WriteLogCR("Error: Class not aggragatable ");
			return -2;
		case E_NOINTERFACE:
			WriteLogCR("Error: Class does not support interface ");
			return -3;
		case E_FAIL :
			WriteLogCR("Error: Failure ");
			return -4;
			break;
		case E_POINTER:
			WriteLogCR("Error: NULL pointer ");
			return -5;
		case E_INVALIDARG:
			WriteLogCR("Error: Invalid argument");
			return -6;
		case E_NOTIMPL :
			WriteLogCR("Error: not implemented");
			return -7;
		default:
			WriteLogNum("Error: Unknown error ",lret,TRUE);
			return -8;
	}

	return -8;
}

static IPin * GetPin(IBaseFilter * basefilter,BOOL bInput,AM_MEDIA_TYPE * lpamtype)
{
	IPin * ipinout = NULL;
	ULONG ulx;
	PIN_INFO pininfo;
	IEnumMediaTypes * ienummt = NULL;
	AM_MEDIA_TYPE  * amt;
	IPin * tmp;
	HRESULT hr;
	IEnumPins * enumpins = NULL;

	if( (basefilter->EnumPins(&enumpins) != S_OK) || (enumpins == NULL) )
	{
		return NULL;
	}
	
	enumpins->Reset();
	ipinout = NULL;
	while(enumpins->Next(1,&ipinout,&ulx) == S_OK)
	{
		if(ipinout == NULL) 
		{
			return NULL;
		}
		
		if(ipinout->QueryPinInfo(&pininfo) != S_OK)
		{
			return NULL;
		}

		hr = ipinout->ConnectedTo(&tmp);
		if( (hr != S_OK) && (pininfo.dir == (bInput ? PINDIR_INPUT : PINDIR_OUTPUT)) )
		{
			if((ipinout->EnumMediaTypes(&ienummt) == S_OK) && (ienummt != NULL))
			{
				if( (ienummt->Reset() == S_OK) && (ienummt->Next(1,&amt,&ulx) == S_OK))
				{
					if(lpamtype != NULL)
					{
						memcpy(lpamtype,amt,sizeof(AM_MEDIA_TYPE));
					}
					break;
				}
			
			}
		}
		
	}

	return ipinout;
}


//Direct show wav - > oma entry
static int RenderOMAFileToWav(LPSTR lpstrSrc,LPSTR lpstrDest)
{
	
	char x;
	long lret = 0;
	DWORD dwRet = 0;
	HRESULT hr;
	LONG levent,lparam1,lparam2;
	BOOL bExit = FALSE;
	BOOL bCheckPos = FALSE;
	WCHAR lpwstrWideChar[MAX_PATH];

	///-----Graph objects----
	IMediaControl * imedia = NULL;
	IMediaSeeking * iseek = NULL;
	IMediaPosition * ipos = NULL;
	IGraphBuilder  * igraph = NULL;
	LONGLONG lllen,llref,llstop,lltmp,llearly,lllatest,lllastpos;
	REFTIME rtstop,rttmp,rtref,rtlen;
	IMediaEvent * ievent = NULL;
	GUID timeformat = TIME_FORMAT_MEDIA_TIME;
	DWORD dwCap = 0;
	//----omgsrc objects----
	IUnknown * omgsrc_unknown = NULL;
	IBaseFilter * omgsrc_base= NULL;
	IPin * omgsrc_ipinout = NULL;
	AM_MEDIA_TYPE omgsrc_pinamt;
	IMediaSample * omgsrc_medsmp = NULL;
	//------omg transform---
	IUnknown * omgtrns_unknown = NULL;
	IBaseFilter * omgtrns_base= NULL;
	IPin * omgtrns_ipinin = NULL;
	IPin * omgtrns_ipinout = NULL;
	AM_MEDIA_TYPE amt_ipinout;
	//-------smart tee------------
	IUnknown * smarttee_unknown = NULL;
	IBaseFilter * smarttee_base = NULL;
	IPin * smarttee_pinin = NULL;
	IPin * smarttee_pinout = NULL;
	IPin * smarttee_pinprev = NULL;
	//------direct sound---------
	IUnknown * dsound_unknown = NULL;
	IBaseFilter * dsound_base= NULL;
	IReferenceClock * dsound_refclock = NULL;
	IPin * dsound_ipinin = NULL;
	//------sony wav writer---------
	IUnknown * sonywav_unknown = NULL;
	IBaseFilter * sonywav_base= NULL;
	IPin * sonywav_ipinin = NULL;
	IFileSinkFilter * sonywav_filesink = NULL;
	
	CLSID clsidwavw = {0x6d6533f6, 0x5968, 0x4fb7, {0x8c, 0x00,  0x90, 0x5e,  0x71,0x57, 0x3d, 0xc4}};
	CLSID clsidomg =  {0x98660581, 0xc9a8, 0x4c92, {0xb4, 0x80,  0xf2, 0x7d , 0xe3,0xc3, 0xaa, 0xb4}};
	
	if(dwRet == 0)
	{
		WriteLog("Create filter graph --",FALSE);
		if(DecodeRet(CoCreateInstance(CLSID_FilterGraph,NULL,CLSCTX_INPROC_SERVER ,IID_IGraphBuilder,(void**)&igraph)) != 0)
		{
			dwRet = 1;
		}
		else if(DecodeRet(igraph->QueryInterface(IID_IMediaControl,(void**)&imedia)) != 0)
		{
			dwRet = 1;
		}
		else if(DecodeRet(igraph->QueryInterface(IID_IMediaPosition ,(void**)&ipos)) != 0)
		{
			dwRet = 1;
		}
		else if(DecodeRet(igraph->QueryInterface(IID_IMediaEvent ,(void**)&ievent)) != 0)
		{
			dwRet = 1;
		}
		
	}

	if(hLogFile != INVALID_HANDLE_VALUE)
	{
		igraph->SetLogFile((DWORD)hLogFile);
	}
	

	//Get IBaseFilter interface from default direct sound device
	if(dwRet == 0)
	{
		WriteLog("Get direct sound filters --",FALSE);
		if(DecodeRet(CoCreateInstance(CLSID_DSoundRender,NULL,CLSCTX_INPROC_SERVER ,IID_IUnknown,(void**)&dsound_unknown)) != 0)
		{
			dwRet = 2;
		}
		else if(DecodeRet(dsound_unknown->QueryInterface(IID_IBaseFilter,(void**)&dsound_base)) != 0)
		{
			dwRet = 2;
		}
		else if(DecodeRet(dsound_base->QueryInterface(IID_IReferenceClock,(void**)&dsound_refclock)) != 0)
		{
			dwRet = 3;
		}
		
	}

	//add direct sound to graph
	if(dwRet == 0)
	{
		WriteLog("add direct sound to graph ",FALSE);
		if (DecodeRet(igraph->AddFilter(dsound_base,L"dsound")) != 0)
		{
			dwRet = 4;
		}
	}

	//get input pin from sound 
	if(dwRet == 0)
	{
		WriteLog("get input pin from dsound ",FALSE);
		dsound_ipinin = GetPin(dsound_base,TRUE,NULL);
		if(dsound_ipinin == NULL)
		{
			dwRet = 5;
		}
	}

	//Get IBaseFilter interface from OMG Source 
	if(dwRet == 0)
	{
		WriteLog("Get omgsrc_base",FALSE);
		MultiByteToWideChar(CP_ACP,0,lpstrSrc,-1,lpwstrWideChar,(MAX_PATH) -1);
		if(DecodeRet(igraph->AddSourceFilter(lpwstrWideChar,L"Src",&omgsrc_base)) != 0)
		{
			dwRet = 5;
		}
		else if(DecodeRet(omgsrc_base->QueryInterface(IID_IUnknown,(void **)&omgsrc_unknown)) != 0)
		{
			dwRet = 5;
		}
	}
		
	//get output pin from OMG Source
	if(dwRet == 0)
	{
		WriteLog("Get omgsrc_base output pin ",FALSE);
		omgsrc_ipinout = GetPin(omgsrc_base,FALSE,&omgsrc_pinamt);
		if(omgsrc_ipinout == NULL)
		{
			WriteLogCR("Failed");
			dwRet = 6;
		}
		else 
		{
			WriteLogCR("Success ");
		}
	}
		
	//create OMG TRANSFORM
	if(dwRet == 0)
	{
		WriteLog("Create OMG TRANSFORM ",FALSE);
		
		if(DecodeRet(CoCreateInstance(clsidomg,NULL,CLSCTX_INPROC_SERVER ,IID_IUnknown,(void**)&omgtrns_unknown)) != 0)
		{
			dwRet = 7;
		}
		else if(DecodeRet(omgtrns_unknown->QueryInterface(IID_IBaseFilter,(void**)&omgtrns_base)) != 0)
		{
			dwRet = 7;
		}
	}

	//add OMG TRNASSFORM to graph
	if(dwRet == 0)
	{
		WriteLog("Add OMG TRANSFORM to graph ",FALSE);
		if(DecodeRet(igraph->AddFilter(omgtrns_base,L"Trns")) != 0)
		{
			dwRet = 8;
		}
	}


	//get input pin from omg trans
	if(dwRet == 0)
	{
		WriteLog("get input in from omg trans ",FALSE);
		if(DecodeRet(omgtrns_base->FindPin(L"Input",&omgtrns_ipinin)) != 0)
		{
			dwRet = 9;
		}
	}

	//get output pin from omg trans
	if(dwRet == 0)
	{
		WriteLog("get output in from omg trans ",FALSE);
		omgtrns_ipinout = GetPin(omgtrns_base,FALSE,&amt_ipinout);
		if(omgtrns_ipinout == NULL)
		{
			dwRet = 10;
			WriteLogCR(" - Error ");
		}
		else
		{
			WriteLogCR(" - Success! ");
		}
	}

	//Get IBaseFilter interface from sony wav writer
	if(dwRet == 0)
	{
		WriteLog("Get sony wav writer ",FALSE);
		if(DecodeRet(CoCreateInstance(clsidwavw,NULL,CLSCTX_INPROC_SERVER ,IID_IUnknown,(void**)&sonywav_unknown)) != 0)
		{
			dwRet = 11;
		}
		else if(DecodeRet(sonywav_unknown->QueryInterface(IID_IBaseFilter,(void **)&sonywav_base)) != 0)
		{
			dwRet = 11;
		}
		else if(DecodeRet(sonywav_unknown->QueryInterface(IID_IFileSinkFilter,(void **)&sonywav_filesink)) != 0)
		{
			dwRet = 11;
		}
	}

	//add sony wav writer to graph
	if(dwRet == 0)
	{
		WriteLog("Add sony wav writer to graph ",FALSE);
		if(DecodeRet(igraph->AddFilter(sonywav_base,L"writer")) != 0)
		{
			dwRet = 12;
		}
	}

	//set up sony wav writer output file
	if(dwRet == 0)
	{
		MultiByteToWideChar(CP_ACP,0,lpstrDest,-1,lpwstrWideChar,255);

		WriteLog("Delete output File: ",FALSE);
		WriteLog(lpstrDest,FALSE);

		if(DeleteFile(lpstrDest))
		{
			WriteLogCR(" -- Success!");
		}
		else
		{
			WriteLogCR(" -- FAILED");
		}
		
		WriteLog("Set up sony wav writer ",FALSE);
		if(DecodeRet(sonywav_filesink->SetFileName(lpwstrWideChar,NULL)) != 0)
		{
			dwRet = 13;
		}		
	}

	//get input pin from sony wav writer
	if(dwRet == 0)
	{
		WriteLog("get input in from omg trans ",FALSE);
		if(DecodeRet(sonywav_base->FindPin(L"Input",&sonywav_ipinin)) != 0)
		{
			dwRet = 14;
		}
	}
	

	//get IBaseFilter interface from smarttee
	if(dwRet == 0)
	{
		WriteLog("Get smarttee interfaces ",FALSE);
		if(DecodeRet(CoCreateInstance(CLSID_SmartTee,NULL,CLSCTX_INPROC_SERVER ,IID_IUnknown,(void**)&smarttee_unknown)) != 0)
		{
			dwRet = 15;
		}
		else if(DecodeRet(smarttee_unknown->QueryInterface(IID_IBaseFilter,(void **)&smarttee_base)) != 0)
		{
			dwRet = 15;
		}	
	}

	//add smarttee to graph
	if(dwRet == 0)
	{
		WriteLog("Add smarttee to graph ",FALSE);
		if(DecodeRet(igraph->AddFilter(smarttee_base,L"smarttee")) != 0)
		{
			dwRet = 16;
		}
	}

	//get smarttee pins
	if(dwRet == 0)
	{
		WriteLog("Get Smarttee pins ",FALSE);
		if(DecodeRet(smarttee_base->FindPin(L"Input",&smarttee_pinin)) != 0)
		{
			dwRet = 17;
		}
		else if(DecodeRet(smarttee_base->FindPin(L"Capture",&smarttee_pinout)) != 0)
		{
			dwRet = 17;
		}
		else if(DecodeRet(smarttee_base->FindPin(L"Preview",&smarttee_pinprev)) != 0)
		{
			dwRet = 17;
		}
	}
	
	//attempt src->trns connection
	if(dwRet == 0)
	{
		WriteLog("Attempt connection between omgsrc and omgtrans ",FALSE);

		if(DecodeRet(igraph->Connect(omgsrc_ipinout,omgtrns_ipinin)) != 0)
		{
 			dwRet = 18;
		}
	}

	//attempt trns->smarttee connection
	if(dwRet == 0)
	{
		WriteLog("Attempt connection between  omgtrans and smarttee ", FALSE);
		if(DecodeRet(igraph->Connect(omgtrns_ipinout,smarttee_pinin)) != 0)
		{
 			dwRet = 19;
		}
	
	}

	//attempt smarttee->wav writer connection
	if(dwRet == 0)
	{
		WriteLog("Attempt connection between  smarttee and sony wav writer ",FALSE);
		if(DecodeRet(igraph->Connect(smarttee_pinout,sonywav_ipinin)) != 0)
		{
 			dwRet = 20;
		}
	
	}

	//connect smarttee and direct sound
	if(dwRet == 0)
	{
		WriteLog("Attempt connection between  smarttee and directsound ",FALSE);
		if(DecodeRet(igraph->Connect(smarttee_pinprev,dsound_ipinin)) != 0)
		{
 			dwRet = 21;
		}
	}

	//set up graph 
	if(dwRet == 0)
	{
		WriteLog("Setting up graph ",FALSE);
		
		if(DecodeRet(igraph->QueryInterface(IID_IMediaSeeking ,(void**)&iseek)) != 0)
		{
			dwRet = 22;
		}
		else if(DecodeRet(iseek->SetTimeFormat(&timeformat)) != 0)
		{
			dwRet = 22;
		}
		else if (DecodeRet(ievent->CancelDefaultHandling(EC_COMPLETE)) != 0)
		{
			dwRet = 23;
		}
		else if(DecodeRet(iseek->GetDuration(&lllen)) != 0)
		{
			dwRet = 23;
		}
		else if(DecodeRet(ipos->get_Duration(&rtlen)) != 0)
		{
			dwRet = 23;
		}

		dwCap = 0xffffffff;
		hr = iseek->GetCapabilities(&dwCap);
		iseek->GetPreroll(&llref);
		WriteLogNum("Capabilities:",dwCap,FALSE);
		WriteLogNumLongLong(" preroll:" ,llref,TRUE);	
	}


	//Run filter
	if(dwRet == 0)
	{
		WriteLogNum("Run. omgsrc= ",(DWORD)omgsrc_base,FALSE);
		WriteLogNum(" omgtrns=", (DWORD)omgtrns_base,FALSE);
		WriteLogNum("smarttee=", (DWORD)smarttee_base,FALSE);
		WriteLogNum("sonywav=", (DWORD)sonywav_base,FALSE);
		WriteLogNum("sound=",(DWORD)dsound_base,TRUE);
			
		rtref=  0;
		llref = 0;
		llstop = dwBlockSize * 10000000;
		hr = iseek->SetPositions(&llref,AM_SEEKING_AbsolutePositioning | AM_SEEKING_Segment ,
				                &llstop,AM_SEEKING_AbsolutePositioning | AM_SEEKING_Segment  );
		if(hr != S_OK)
		{
			WriteLogNum("Error setting end ",hr,TRUE);
			ipos->put_StopTime(rtref + dwBlockSize);
		}


		imedia->Run();
		lllastpos = -1;
		do
		{
			levent = 0;
			lparam2 = 0;
			lparam1 = 0;
			ievent->GetEvent(&levent,&lparam1,&lparam2,INFINITE);
			switch(levent)
			{
				case EC_PAUSED:		
					
					break;
				case EC_COMPLETE:
					hr = iseek->GetCurrentPosition(&lltmp);
					if(hr != S_OK)
					{
						WriteLogNum("Error getting fine pos ",hr,TRUE);
						ipos->get_CurrentPosition(&rttmp);
						lltmp = rttmp * 10000000;
					}
					

					if( ((hr == S_OK) && (lltmp >= lllen)) ||
						(lllastpos == lltmp)	//infinite loop detection
					  )
					{
						bExit = TRUE;
					}
					else
					{	
						//pause and read current position
						imedia->Pause();
						hr = iseek->GetCurrentPosition(&lltmp);
						WriteLogNumLongLong("Stopped at ",lltmp,TRUE);

						//set resume position
						rtref += dwBlockSize - 1;
						llref +=  (dwBlockSize * 10000000) - (dwBlockSize * 1000000);
						llstop = llref + ((dwBlockSize * 10000000) + (dwBlockSize * 1000000));
						lllastpos = lltmp;
			
						if(llstop >= lllen)
						{	//we're near the end... Lets just run to the end. without a stop position
																			
							hr = iseek->SetPositions(&llref,AM_SEEKING_AbsolutePositioning | AM_SEEKING_NoFlush ,
													 &llstop,AM_SEEKING_NoPositioning | AM_SEEKING_NoFlush);
						}
						else
						{
							hr = iseek->SetPositions(&llref,AM_SEEKING_AbsolutePositioning | AM_SEEKING_NoFlush ,
													 &llstop,AM_SEEKING_AbsolutePositioning | AM_SEEKING_NoFlush);
						}


						if(hr != S_OK)
						{
							WriteLogNum("error seeking ",hr,TRUE);
							ipos->put_CurrentPosition(rtref - 1);
							ipos->put_StopTime(rtref + dwBlockSize);
						}

						//resume filter graph
						imedia->Run();						
					}
					
					break;
			}
							
			ipos->get_CurrentPosition(&rttmp);
			WriteLogNum("Pos: ",(DWORD)rttmp,FALSE);
			WriteLogNum("Len: ",(DWORD)rtlen,FALSE);
			WriteLogNumLongLong("Pos (secs): ",lltmp,FALSE);
			WriteLogNumLongLong("Len (secs): ",lllen,FALSE);
			WriteLogNum("Event: ",levent,TRUE);
		}
		while(bExit == FALSE);
	}

	
	//Release everything
	///-----Graph objects----
	RELEASE(imedia);
	RELEASE(iseek);
	RELEASE(ipos);
	RELEASE(igraph);
	RELEASE(ievent);
	//----omgsrc objects----
	RELEASE(omgsrc_unknown);
	RELEASE(omgsrc_base);
	RELEASE(omgsrc_ipinout);
	RELEASE(omgsrc_medsmp);
	//------omg transform---
	RELEASE(omgtrns_unknown);
	RELEASE(omgtrns_base);
	RELEASE(omgtrns_ipinin);
	RELEASE(omgtrns_ipinout);
	//-------smart tee------------
	RELEASE(smarttee_unknown);
	RELEASE(smarttee_base);
	RELEASE(smarttee_pinin);
	RELEASE(smarttee_pinout);
	RELEASE(smarttee_pinprev );
	//------direct sound---------
	RELEASE(dsound_unknown);
	RELEASE(dsound_base);
	RELEASE(dsound_refclock);
	RELEASE(dsound_ipinin);
	//------sony wav writer---------
	RELEASE(sonywav_unknown);
	RELEASE(sonywav_base);
	RELEASE(sonywav_ipinin);
	RELEASE(sonywav_filesink);

	return 0;
}

//==================================WAV fixing functions===============================

static BOOL __inline CompareSamples(DWORD dwSample1,DWORD dwSample2)
{
	DWORD a,b,dwDif;
	
	if(dwSample1 != dwSample2 )
	{	
		//not equal... but very small differences are allowed
		a = dwSample1 & 0xffff;
		b = dwSample2 & 0xffff;
		dwDif = (a > b) ? a - b : b - a;
		if((dwDif > dwAllowedDiff))
		{
			return FALSE;
		}

		a = dwSample1 >> 16;
		b = dwSample2 >> 16;
		dwDif = (a > b) ? a - b : b - a;
		if((dwDif > dwAllowedDiff))
		{
			return FALSE;
		}
	}

	return TRUE;
}

static DWORD __inline FindStartOfRepeat(LPBYTE lpbtWavData, DWORD dwSectionOff,
										DWORD dwSearchOff, DWORD  dwRptSrchOff)
{
	LPDWORD lpdwdSrch = (LPDWORD) &lpbtWavData[dwSearchOff + dwSectionOff]; 
	LPDWORD lpdwdRpt = (LPDWORD) &lpbtWavData[dwRptSrchOff + dwSectionOff];
	DWORD dwSub = 0;

	while(CompareSamples(*lpdwdSrch,*lpdwdRpt))
	{
		lpdwdRpt--;
		lpdwdSrch--;
		dwSub += 4;
	}

	return dwSub;
	
}

static DWORD __inline FindLenOfRepeat(LPBYTE lpbtWavData,DWORD dwWavLen,  DWORD dwSectionOff, DWORD dwMaxRptSzInBytes,
									  DWORD dwSearchOff, DWORD  dwRptSrchOff)
{
	DWORD dwRptLen;
	DWORD dwDif;
	WORD a,b;
	LPDWORD lpdwdSrch = (LPDWORD) &lpbtWavData[dwSearchOff + dwSectionOff + 4]; 
	LPDWORD lpdwdRpt = (LPDWORD) &lpbtWavData[dwRptSrchOff + dwSectionOff + 4];

	//now scan for end of repeat
	dwRptLen = 4;
	while( (dwRptLen < dwMaxRptSzInBytes) && 
		   ((dwSectionOff + dwRptLen + dwRptSrchOff) < dwWavLen)
		 )
	{
		if(!CompareSamples(*lpdwdSrch,*lpdwdRpt))
		{
			CompareSamples(*lpdwdSrch,*lpdwdRpt);
			
			return dwRptLen;
		}
		
		lpdwdSrch++;
		lpdwdRpt++;
		dwRptLen += 4;
	}	

	WriteLogNum("Ignoring Repeat as it is longer than maximum",dwMaxRptSzInBytes,TRUE);
	return 0;
}


static BOOL __inline FindRepeatInSection(LPBYTE lpbtWavData, DWORD dwWavLen, DWORD dwSectOff,DWORD dwSectionLen,
										 DWORD dwMinRptSzInBytes,DWORD dwMaxRptSzInBytes,
										 DWORD * lpdwRptOff, DWORD * lpdwRptLen)
{
	DWORD dwRptLen ;
	DWORD dwSearchOff;
	DWORD dwRptSrchOff;
	DWORD dwSmp,dwCmpSmp;
	DWORD dwSub;
	DWORD dwAmount = dwSectionLen / 2;
	DWORD i;	

	
	for(i=0; i < 3;i++)
	{
		WriteLogNum("Using search method ",i,TRUE);
		switch(i)
		{
			case 0:
				dwAmount = dwSectionLen / 2;
				dwSearchOff = dwAmount;
				break;
			case 1:
				dwAmount = dwSectionLen / 2;
				dwSearchOff = dwAmount;
				break;
			case 2:
				dwSearchOff = 8;
				break;
		}
			
		while( (dwSearchOff >= 8) && (dwSearchOff < (dwSectionLen - dwMinRptSzInBytes)) && 
			   ((dwSearchOff+ dwSectOff) < dwWavLen)
			)
		{
			dwSearchOff -= (dwSearchOff % 4);
			dwSmp = *(DWORD *)&lpbtWavData[dwSectOff + dwSearchOff];

			//search for repeat of sample
			dwRptSrchOff = dwSearchOff + 4;
			while( (dwRptSrchOff < dwSectionLen) && 
				   ((dwRptSrchOff + dwSectOff) < dwWavLen)
				 )
			{
				dwCmpSmp = *(DWORD *)&lpbtWavData[dwSectOff + dwRptSrchOff];
				if(CompareSamples(dwSmp,dwCmpSmp))
				{	//match... but is it long enough
					
					dwRptLen = FindLenOfRepeat(lpbtWavData,dwWavLen, 
											   dwSectOff,dwMaxRptSzInBytes,
											   dwSearchOff,dwRptSrchOff);

					if(dwRptLen == 0)
					{	//scanned from dwSectOff+4 to dwMaxRptSzInBytes and nothing found
						dwRptSrchOff += dwMaxRptSzInBytes;
					}
					else if((dwRptLen >= dwMinRptSzInBytes)   && (dwRptLen < dwMaxRptSzInBytes))
					{
						//find start of repeat and return TRUE
						dwSub = FindStartOfRepeat(lpbtWavData,dwSectOff,dwSearchOff,dwRptSrchOff);
						*lpdwRptOff = dwSearchOff - dwSub;
						*lpdwRptLen = dwRptSrchOff  -  dwSearchOff;

						return TRUE;
					}
					else
					{	//repeat block is too small... try again
						dwRptSrchOff += dwRptLen + 4;
					}

				}
				else
				{
					dwRptSrchOff += 4;
				}
			}
			
			switch(i)
			{
				case 0:
					dwAmount /= 2;
					dwSearchOff = dwAmount;
					break;
				case 1:
					dwAmount /= 2;
					dwSearchOff = (dwSectionLen - dwAmount);
					break;
				case 2:
					dwSearchOff += dwMinRptSzInBytes / 2;
					break;
			}
		}
	}

	//not found
	return FALSE;
}

static LPCUTINFO ScanWave(LPBYTE lpbtWavData,DWORD dwWavDataLen,WAVFORMAT * lpwavformat,
						   double dSectionDistInSecs,double dSectionSizeInSecs,
						   double dMinRepeatSizeInSecs,double dMaxRepeatSizeInSecs)
{
	DWORD dwCurSection;
	DWORD dwSectionPosInBytes,dwSectionSizeInBytes,dwMinRptSzInBytes,dwMaxRptSzInBytes;
	DWORD dwChnsBytesPerSample;
	DWORD dwRptOff,dwRptLen;
	DWORD dwMaxCuts,dwCutIndex;
	double x,y,z,dist,avgrptoff,avgrptlen;
	LPCUTINFO lpcutinfo;

	//init
	avgrptoff = 0;
	avgrptlen = 0;
	dwCurSection = 1;
	dwSectionPosInBytes = 0;
	dwChnsBytesPerSample = (lpwavformat->wChannels * (lpwavformat->wBitsPerSample / 8));
	
	//calculate Section size, in bytes
	x = (dSectionSizeInSecs * lpwavformat->dwSampleRate * dwChnsBytesPerSample);
	dwSectionSizeInBytes = x;
	
	//calculate Section distance, in bytes
	dist= dSectionDistInSecs * lpwavformat->dwSampleRate * dwChnsBytesPerSample;
	dwSectionPosInBytes = dist;
	dwSectionPosInBytes -= dwSectionSizeInBytes / 2;
	 		
	//calculate minimum allowed size of repeat data, in bytes
	x = dMinRepeatSizeInSecs * lpwavformat->dwSampleRate * dwChnsBytesPerSample;
	dwMinRptSzInBytes = x;

	//calculate maximum allowed size of repeat data, in bytes
	x = dMaxRepeatSizeInSecs * lpwavformat->dwSampleRate * dwChnsBytesPerSample;
	dwMaxRptSzInBytes = x;

	//calculate maximum number of cut informations that need storing
	x = dwWavDataLen / dist;
	dwMaxCuts = x + 10;
	dwCutIndex = 0;
	lpcutinfo = (LPCUTINFO)LocalAlloc(LPTR,dwMaxCuts * sizeof(CUTINFO));
	memset(lpcutinfo ,0 ,dwMaxCuts * sizeof(CUTINFO));
	
	while(dwSectionPosInBytes < dwWavDataLen)
	{
		x = ((dwSectionPosInBytes + dwSectionSizeInBytes) / dwChnsBytesPerSample) / lpwavformat->dwSampleRate;
		y = (dwSectionPosInBytes / dwChnsBytesPerSample) / lpwavformat->dwSampleRate;
		
		WriteLogNum("SECTION:",dwCurSection,FALSE);
		WriteLogNum("Pos:",dwSectionPosInBytes,FALSE);
		WriteLogNum("Len:",dwWavDataLen, FALSE);
		WriteLogNum("Current section:",dwSectionPosInBytes,FALSE);
		WriteLogNum("Current section size:",dwSectionSizeInBytes,FALSE);
		WriteLogNumDouble("Avg rpt offset:",avgrptoff / dwCurSection,FALSE);
	//	WriteLogNumDouble("Avg rpt length:",avgrptlen / dwCurSection,FALSE);
		WriteLogNumDouble("Section from ",y,FALSE);
		WriteLogNumDouble("to ",x,TRUE);

		if( (dwSectionPosInBytes + dwSectionSizeInBytes) > dwWavDataLen)
		{
			dwSectionSizeInBytes = (dwSectionPosInBytes + dwSectionSizeInBytes) - dwWavDataLen;
			if((dwMaxRptSzInBytes + dwSectionSizeInBytes) > dwWavDataLen)
			{
				dwMaxRptSzInBytes = (dwSectionPosInBytes + dwSectionSizeInBytes) - dwWavDataLen;
			}
		}
		
		//is there repeated audio in this section?
		if(FindRepeatInSection(lpbtWavData ,dwWavDataLen, dwSectionPosInBytes,dwSectionSizeInBytes,
								dwMinRptSzInBytes,dwMaxRptSzInBytes,
								&dwRptOff,&dwRptLen) == TRUE)
		{	
			//found a repeat in the specified section
			x = dwRptLen;
			x /=  dwChnsBytesPerSample * lpwavformat->dwSampleRate;
			WriteLogNum("   Found repeat -> Offset",dwRptOff,FALSE);
			WriteLogNum(" Length:",dwRptLen,FALSE);
			WriteLogNumDouble(" len in secs:",x,TRUE);


			//set up cut info
			lpcutinfo[dwCutIndex].lpbtCutFrom = lpbtWavData + dwSectionPosInBytes + dwRptOff;
			lpcutinfo[dwCutIndex].dwCutLen = dwRptLen;
			dwCutIndex++;

			if(!bUseAvging)
			{
				dwSectionPosInBytes += dwRptOff - dwRptLen;

				//keep stats up to date even though we dont use them
				dwCurSection++;
				avgrptoff += dwRptOff;
			}
			else if( (int)dwRptOff > 0)
			{
				avgrptoff += dwRptOff;
				dwSectionPosInBytes += avgrptoff / dwCurSection ;
				dwCurSection++;
			}
			else
			{
				avgrptoff += (int)dwRptOff;
				dwSectionPosInBytes += dwRptOff;
				dwCurSection++;
			}
			
			
		}
		else
		{	//no
			WriteLogCR("    NOT Found repeat ");
			
		}

		dwSectionPosInBytes += dist;
							
	}

	return lpcutinfo;
}


static DWORD CalculateCutInfoSize(LPCUTINFO lpcutinfo)
{
	DWORD dwRet = 0;
	while(lpcutinfo->lpbtCutFrom != NULL)
	{
		dwRet += lpcutinfo->dwCutLen;
		lpcutinfo++;
	}

	return dwRet;
}

//repair entry function
static BOOL RepairWav(LPSTR lpstrWavFile,LPSTR outfile,double dSectionDistInSecs,double dSectionSizeInSecs,
					  double dMinRepeatSizeInSecs,double dMaxRepeatSizeInSecs)
{
	HANDLE hFile,hFileOut,hmap;
	LPBYTE lpbtWavFile;
	LPBYTE lpbtWavData;
	LPBYTE lpbtTmp;
	WAVSTRUCT * lpwavstruct;
	WAVCHUNK * lpwavchunk;
	WAVFORMAT * lpwavformat;
	DWORD dwFileSize;
	DWORD dwWavDataLen;
	LPCUTINFO lpcutinfo;
	char buffer[255];
	DWORD dwWrit;
	DWORD dwValue;
	DWORD i;
	LPBYTE lpbtFrom;
	
	hFile = CreateFile(lpstrWavFile,GENERIC_READ | GENERIC_WRITE,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
	if(hFile == INVALID_HANDLE_VALUE)
	{
		WriteLog("RepairWav => Failed to open source wave ",FALSE);
		WriteLogCR(lpstrWavFile);
		return FALSE;
	}

	hFileOut = CreateFile(outfile,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,0);
	if(hFileOut == INVALID_HANDLE_VALUE)
	{
		WriteLog("RepairWav => Failed to open dest wave ",FALSE);
		WriteLogCR(outfile);
		CloseHandle(hFile);
		return FALSE;
	}

	dwFileSize = GetFileSize(hFile,NULL);

	hmap = CreateFileMapping(hFile,NULL,PAGE_READONLY,0,0,NULL);
	if(hmap == NULL)
	{
		WriteLog("RepairWav => Failed to create mapped file ",FALSE);
		WriteLogCR(lpstrWavFile);
		CloseHandle(hFile);
		CloseHandle(hFileOut);
		return FALSE;
	}

	lpbtWavFile = (LPBYTE)MapViewOfFile(hmap,FILE_MAP_READ,0,0,0);
	if(lpbtWavFile == NULL)
	{
		WriteLog("RepairWav => Failed map view of file ",FALSE);
		WriteLogCR(lpstrWavFile);
		CloseHandle(hmap);
		CloseHandle(hFile);
		CloseHandle(hFileOut);
		return FALSE;
	}

	//parse wav file header
	lpwavstruct = (WAVSTRUCT *)lpbtWavFile;
	lpwavchunk = (WAVCHUNK *)&lpbtWavFile[sizeof(WAVSTRUCT)];
	while(lpwavchunk != NULL)
	{
		lpbtTmp = (LPBYTE)lpwavchunk;
		if( (lpwavchunk->chunkid[0] == 'f') && (lpwavchunk->chunkid[1] == 'm') &&
			(lpwavchunk->chunkid[2] == 't') )
		{	//fmt chunk.. copy useful bits off
			lpwavformat = (WAVFORMAT *)(lpbtTmp + sizeof(WAVCHUNK));
			if(lpwavformat->wFormatTag != 0x0001)
			{
				lpwavformat = NULL;
				WriteLogCR("Wav Format not PCM ");
				break;
			}

			lpbtTmp += sizeof(WAVFORMAT) +  sizeof(WAVCHUNK);
			lpwavchunk = (WAVCHUNK *)lpbtTmp;
		}
		else if( (lpwavchunk->chunkid[0] == 'd') && (lpwavchunk->chunkid[1] == 'a') &&
			     (lpwavchunk->chunkid[2] == 't') )
		{	//data chunk
			lpbtWavData = (LPBYTE)lpwavchunk;
			dwWavDataLen = lpwavchunk->dwChunkDataLen;
			if(dwWavDataLen == 0)
			{//guess

				dwWavDataLen += (DWORD)lpbtWavData;
				dwWavDataLen -= (DWORD)lpbtWavFile;
				dwWavDataLen = dwFileSize - dwWavDataLen;
				dwWavDataLen -= sizeof(WAVCHUNK);
			}


			lpbtWavData += sizeof(WAVCHUNK);
			break;
		}
		else
		{	//invalid
			break;
		}
		
	}
	
	//check all required wave components are there....
	if( (lpwavchunk == NULL)  || (lpbtWavData == NULL) )
	{
		UnmapViewOfFile(lpbtWavFile);
		lpbtWavFile = NULL;
		CloseHandle(hmap);
		CloseHandle(hFileOut);
		CloseHandle(hFile);
		WriteLogCR("Missing data in wave header");
		return FALSE;
	}

	//scan wave
	lpcutinfo = NULL;
	lpcutinfo = ScanWave(lpbtWavData,dwWavDataLen,lpwavformat,
						 dSectionDistInSecs,dSectionSizeInSecs,
						 dMinRepeatSizeInSecs,dMaxRepeatSizeInSecs);

	//------Create New Wav----------
	//write RIFF header
	sprintf(buffer,"RIFF");	
	dwValue = ((dwWavDataLen - CalculateCutInfoSize(lpcutinfo)) - 84);		
	WriteFile(hFileOut,buffer,4,&dwWrit,NULL);	//
	WriteFile(hFileOut,&dwValue,4,&dwWrit,NULL);
	sprintf(buffer,"WAVE");
	WriteFile(hFileOut,buffer,4,&dwWrit,NULL);

	//write format chunk
	sprintf(buffer,"fmt ");
	WriteFile(hFileOut,buffer,4,&dwWrit,NULL);	//chunk id
	dwValue = 16;
	WriteFile(hFileOut,&dwValue,4,&dwWrit,NULL);	//chunk data size
	WriteFile(hFileOut,&lpwavformat->wFormatTag,2,&dwWrit,NULL);	//compression code
	WriteFile(hFileOut,&lpwavformat->wChannels,2,&dwWrit,NULL);	//num channels
	WriteFile(hFileOut,&lpwavformat->dwSampleRate,4,&dwWrit,NULL);	//sample rate
	WriteFile(hFileOut,&lpwavformat->dwBufferEst,4,&dwWrit,NULL); //average bytes per second
	WriteFile(hFileOut,&lpwavformat->wBlockAlign,2,&dwWrit,NULL); //block align
	WriteFile(hFileOut,&lpwavformat->wBitsPerSample,2,&dwWrit,NULL); //bits per sample
	dwValue = 0;
	
	//write data chunk
	sprintf(buffer,"data");
	WriteFile(hFileOut,buffer,4,&dwWrit,NULL);	//chunk id
	dwValue = dwWavDataLen - CalculateCutInfoSize(lpcutinfo);
	WriteFile(hFileOut,&dwValue,4,&dwWrit,NULL);	//new wave data len
	
	i=0;
	lpbtFrom = lpbtWavData;
	while(lpbtFrom < (lpbtWavData + dwWavDataLen))
	{
		//copy to beginning of current cut start position
		if(lpcutinfo[i].lpbtCutFrom != NULL)
		{
			WriteFile(hFileOut,lpbtFrom,lpcutinfo[i].lpbtCutFrom - lpbtFrom - 4,&dwWrit,NULL);

			//set new from pointer 
			lpbtFrom = lpcutinfo[i].lpbtCutFrom + lpcutinfo[i].dwCutLen ;
			i++;
		}
		else
		{	//finished... just copy from current pos to end of data
			WriteFile(hFileOut,lpbtFrom,(lpbtWavData + dwWavDataLen) - lpbtFrom,&dwWrit,NULL);
			break;
		}
	}
		

	//clean up
	if(lpcutinfo != NULL)
	{
		LocalFree(lpcutinfo);
	}
 	UnmapViewOfFile(lpbtWavFile);
	lpbtWavFile = NULL;
	CloseHandle(hmap);
	CloseHandle(hFile);
	return TRUE;
}



//========================Main function==============================
static LPSTR FindFirstNonSpace(char * x)
{
	
	if(x == NULL)
	{
		return 0;
	}
	
	while((x[0] == ' ') && (x[0] != 0))
	{
		x++;
	}

	return x;
}

static BOOL CheckFileIsOK(LPSTR name,BOOL bNewFile)
{
	HANDLE hFile = CreateFile(name, 
							  bNewFile ? GENERIC_READ | GENERIC_WRITE : GENERIC_READ ,
							  bNewFile ? 0 : FILE_SHARE_READ,
							  NULL,
							  bNewFile ? CREATE_ALWAYS : OPEN_EXISTING,
							  FILE_ATTRIBUTE_NORMAL,NULL);

	if(hFile != INVALID_HANDLE_VALUE)
	{
		CloseHandle(hFile);
		return TRUE;
	}

	return FALSE;
}

static BOOL ProcessArg(char * x,char * y)
{
	char * param;

	if((x[0] == '-'))
	{	
		switch(x[1])
		{
			case 'i':	
				param = FindFirstNonSpace(x+2);
				if((param == NULL) || (param[0] == 0))
				{
					param = FindFirstNonSpace(y);	
				}

				if(param != 0)
				{
					strcpy(infile,param);
				}

				if(!CheckFileIsOK(infile,FALSE))
				{
					printf("ERROR: Could not open input file %s \n",infile);
					return FALSE;
				}
				break;	
			case 'o':		
				param = FindFirstNonSpace(x+2);
				if((param == NULL) || (param[0] == 0))
				{
					param = FindFirstNonSpace(y);	
				}

				if(param != 0)
				{
					strcpy(destfile,param);
				}

				if(!CheckFileIsOK(destfile,TRUE))
				{
					printf("ERROR: Could not open destination file %s \n",destfile);
					return FALSE;
				}
				break;
			case 't':
				param = FindFirstNonSpace(x+2);
				if((param == NULL) || (param[0] == 0))
				{
					param = FindFirstNonSpace(y);	
				}

				if(param != 0)
				{
					strcpy(tempfile,param);
				}

				if(!CheckFileIsOK(tempfile,TRUE))
				{
					printf("ERROR: Could not open temp file %s \n",tempfile);
					return FALSE;
				}
				break;
			case 'k':
				bKeepTmpFile = TRUE;
				break;
			case 'd':
				bDebug = TRUE;
				break;
			case 'c':
				bUseAvging = FALSE;
				break;
			case 'f':
				param = FindFirstNonSpace(x+2);
				if((param == NULL) || (param[0] == 0))
				{
					param = FindFirstNonSpace(y);	
				}

				if(param == NULL)
				{
					printf("ERROR: Invalid specified maximum difference value ");
					return FALSE;
				}

				dwAllowedDiff = strtol(param,NULL,10);
				break;
			case 'b':
				param = FindFirstNonSpace(x+2);
				if((param == NULL) || (param[0] == 0))
				{
					param = FindFirstNonSpace(y);	
				}
				
				dwBlockSize = strtol(param,NULL,10);

				if(dwBlockSize < 2)
				{
					printf("ERROR: Invalid specified block size");
					return FALSE;
				}
				break;
			default:
				printf("Invalid parameter!\n");
			case 'h':
				printf("===============================================================================\n");
				printf("|               MarC's dodgy OMA -> WAV converter ... v0.02                   |\n");
				printf("===============================================================================\n");
				printf("  -h              Displays this help\n");
				printf("  -i <file>       Specifies input file \n");
				printf("  -o <file>       Specifies output file [OPTIONAL] \n");
				printf("  -t <file>       Specifies temporary file [OPTIONAL] \n");
				printf("  -k              Keep temporary file [OPTIONAL] \n");
				printf("  -b <block size> Read block size in seconds [EXPERIMENTAL] default=10 \n");
				printf("  -f <max value>  Max allowed difference between samples [OPTIONAL] default=15\n");
				printf("  -d              Produce a log file full of debug \n");
				printf("  -c              Use old non-averaging method of offset calculation [OPTIONAL]\n");
				printf("===============================================================================\n");
				return FALSE;
		}
	}

	return TRUE;
}


int main(int argv,char ** argc)
{
	int iRet = -1;
	char x;
	int i;
	
//	return testcode();

	//init
	memset(infile,0,sizeof(infile));
	memset(tempfile,0,sizeof(tempfile));
	memset(destfile,0,sizeof(destfile));

	//process command params
	for(i=1;i < argv;i++)
	{
		if(i != (argv-1))
		{
			if(!ProcessArg(argc[i],argc[i+1]))
			{
				return 0;
			}
		}
		else
		{
			if(!ProcessArg(argc[i],NULL))
			{
				return 0;
			}
		}
	}

	if(infile[0] == 0)
	{
		printf("ERROR: No input file!\n");
		ProcessArg("-h",NULL);
		printf("\n\nPress any key to exit");
		getch();
		return -1;
	}
	
	//more init
	CoInitialize(NULL);
	if(bDebug)
	{
		CreateLogFile(infile);
	}

	WriteLogNum("Specified Block Size: ",dwBlockSize,TRUE);
	
	
	if(tempfile[0] == 0)
	{
		//apend .wav.tmp for temporary file
		sprintf(tempfile,"%s.wav.tmp",infile);
	}

	if(destfile[0] == 0)
	{
		//append .wav for output wav
		sprintf(destfile,"%s.wav",infile);
	}


	iRet = RenderOMAFileToWav(infile,tempfile);
	iRet = 0;

	if(iRet == 0)
	{
		RepairWav(tempfile,destfile,dwBlockSize, (double)dwBlockSize / 2, 0.1,(double)dwBlockSize / 5);
	}

	if(!bKeepTmpFile)
	{
		DeleteFile(tempfile);
	}

	printf("\n\nResult:%u Press anything to exit",iRet);
	getch();

}
